# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import importlib
from abc import (
    ABCMeta,
    abstractmethod
)
from typing import Dict

from kiwi.container.base import ContainerImageBase
from kiwi.exceptions import (
    KiwiContainerImageSetupError
)


class ContainerImage(metaclass=ABCMeta):
    """
    **Container Image factory**

    :param string name: container system name
    :param string root_dir: root directory path name
    :param dict custom_args: custom arguments
    """
    @abstractmethod
    def __init__(self) -> None:
        return None  # pragma: no cover

    @staticmethod
    def new(name: str, root_dir: str, custom_args: Dict=None) -> ContainerImageBase:  # noqa: E252
        name_map = {
            'docker': 'OCI',
            'oci': 'OCI',
            'appx': 'Appx'
        }
        args_map = {
            'docker': [root_dir, 'docker-archive', custom_args],
            'oci': [root_dir, 'oci-archive', custom_args],
            'appx': [root_dir, custom_args]
        }
        try:
            container_image = importlib.import_module(
                'kiwi.container.{}'.format('oci' if name == 'docker' else name)
            )
            module_name = 'ContainerImage{}'.format(name_map[name])
            return container_image.__dict__[module_name](*args_map[name])
        except Exception:
            raise KiwiContainerImageSetupError(
                'Support for {0} container not implemented'.format(name)
            )



from unittest.mock import patch

import unittest.mock as mock

from kiwi.filesystem.fat16 import FileSystemFat16


class TestFileSystemFat16:
    @patch('os.path.exists')
    def setup(self, mock_exists):
        mock_exists.return_value = True
        provider = mock.Mock()
        provider.get_device = mock.Mock(
            return_value='/dev/foo'
        )
        self.fat16 = FileSystemFat16(provider, 'root_dir')
        self.fat16.setup_mountpoint = mock.Mock(
            return_value='some-mount-point'
        )

    @patch('os.path.exists')
    def setup_method(self, cls, mock_exists):
        self.setup()

    @patch('kiwi.filesystem.fat16.Command.run')
    def test_create_on_device(self, mock_command):
        self.fat16.create_on_device('label', 100, uuid='uuid')
        call = mock_command.call_args_list[0]
        assert mock_command.call_args_list[0] == call(
            [
                'mkdosfs', '-F16', '-I', '-n', 'label',
                '-i', 'uuid', '/dev/foo', '100'
            ]
        )

    @patch('kiwi.filesystem.fat16.Command.run')
    def test_set_uuid(self, mock_command):
        self.fat16.set_uuid()
        mock_command.assert_called_once_with(
            ['mlabel', '-n', '-i', '/dev/foo', '::']
        )



from unittest.mock import patch
from pytest import raises

from kiwi.privileges import Privileges

from kiwi.exceptions import KiwiPrivilegesError


class TestPrivileges:
    @patch('os.geteuid')
    def test_check_for_root_permiossion_false(self, mock_euid):
        mock_euid.return_value = 1
        with raises(KiwiPrivilegesError):
            Privileges.check_for_root_permissions()

    @patch('os.geteuid')
    def test_check_for_root_permiossion_true(self, mock_euid):
        mock_euid.return_value = 0
        assert Privileges.check_for_root_permissions() is True



from collections import namedtuple
import logging

from kiwi.logger_filter import (
    LoggerSchedulerFilter,
    InfoFilter,
    DebugFilter,
    ErrorFilter,
    WarningFilter
)


class TestLoggerSchedulerFilter:
    def setup(self):
        self.scheduler_filter = LoggerSchedulerFilter()

    def setup_method(self, cls):
        self.setup()

    def test_filter(self):
        MyRecord = namedtuple(
            'MyRecord',
            'name'
        )
        ignorables = [
            'apscheduler.scheduler',
            'apscheduler.executors.default'
        ]
        for ignorable in ignorables:
            record = MyRecord(name=ignorable)
            assert self.scheduler_filter.filter(record) is False


class TestInfoFilter:
    def setup(self):
        self.info_filter = InfoFilter()

    def setup_method(self, cls):
        self.setup()

    def test_filter(self):
        MyRecord = namedtuple(
            'MyRecord',
            'levelno'
        )
        record = MyRecord(levelno=logging.INFO)
        assert self.info_filter.filter(record) is True


class TestDebugFilter:
    def setup(self):
        self.debug_filter = DebugFilter()

    def setup_method(self, cls):
        self.setup()

    def test_filter(self):
        MyRecord = namedtuple(
            'MyRecord',
            'levelno'
        )
        record = MyRecord(levelno=logging.DEBUG)
        assert self.debug_filter.filter(record) is True


class TestErrorFilter:
    def setup(self):
        self.error_filter = ErrorFilter()

    def setup_method(self, cls):
        self.setup()

    def test_filter(self):
        MyRecord = namedtuple(
            'MyRecord',
            'levelno'
        )
        record = MyRecord(levelno=logging.ERROR)
        assert self.error_filter.filter(record) is True


class TestWarningFilter:
    def setup(self):
        self.error_filter = WarningFilter()

    def setup_method(self, cls):
        self.setup()

    def test_filter(self):
        MyRecord = namedtuple(
            'MyRecord',
            'levelno'
        )
        record = MyRecord(levelno=logging.WARNING)
        assert self.error_filter.filter(record) is True



#!/usr/bin/python3

from textwrap import dedent

import subprocess
import re

import collections


class AutoVivification(dict):
    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value


class AppHash:
    def __init__(self):  # noqa: C901
        tasks = [
            'kiwi/cli.py',
            'kiwi/tasks/*.py'
        ]
        call_parm = ['bash', '-c', 'cat %s' % ' '.join(tasks)]
        cmd = subprocess.Popen(
            call_parm, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        begin = False
        cur_path = ''
        self.result = AutoVivification()
        for line in cmd.communicate()[0].decode().split('\n'):
            usage = re.search('^usage: (.*)', line)
            if usage:
                begin = True
                if re.match('.*--help', line):
                    mod_line = re.sub('[\[\]\|]', '', usage.group(1))
                    mod_line = re.sub('-h ', '', mod_line)
                    key_list = mod_line.split()
                    result_keys = self.validate(key_list)
                    cur_path = self.merge(result_keys, self.result)
                else:
                    key_list = usage.group(1).split()
                    result_keys = self.validate(key_list)
                    cur_path = self.merge(result_keys, self.result)
            elif begin:
                if not line:
                    begin = False
                else:
                    if re.match('.*--version', line):
                        mod_line = re.sub('[\[\]\|]', '', line)
                        mod_line = re.sub('-v ', '', mod_line)
                        key_list = mod_line.split()
                        result_keys = self.validate(key_list)
                        cur_path = self.merge(result_keys, self.result)
                    elif re.match('.*kiwi-ng --compat', line):
                        mod_line = re.sub('[\[\]\|]', '', line)
                        mod_line = re.sub('<legacy_args>...', '', mod_line)
                        key_list = mod_line.split()
                        result_keys = self.validate(key_list)
                        cur_path = self.merge(result_keys, self.result)
                    elif re.match('.*kiwi-ng \[', line):
                        line = line.replace('[', '')
                        line = line.replace(']', '')
                        line = line.replace('|', '')
                        key_list = line.split()
                        key_list.pop(0)
                        for global_opt in key_list:
                            result_keys = self.validate(
                                ['kiwi-ng', global_opt]
                            )
                            cur_path = self.merge(result_keys, self.result)
                    elif re.match('            \[', line):
                        line = line.replace('[', '')
                        line = line.replace(']', '')
                        line = line.replace('|', '')
                        key_list = line.split()
                        for global_opt in key_list:
                            result_keys = self.validate(
                                ['kiwi-ng', global_opt]
                            )
                            cur_path = self.merge(result_keys, self.result)
                    elif re.match('                \[', line):
                        opt_val = re.search('                \[(--.*)', line)
                        global_opt = opt_val.group(1)
                        global_opt = global_opt.replace('[', '')
                        global_opt = global_opt.replace(']', '')
                        global_opt = global_opt.replace('|', '')
                        result_keys = self.validate(
                            ['kiwi-ng', global_opt]
                        )
                        cur_path = self.merge(result_keys, self.result)
                    elif re.match('.*kiwi', line):
                        mandatory_options = re.search(
                            '(.*kiwi-ng.*?) (--.*)', line
                        )
                        if mandatory_options:
                            line = mandatory_options.group(1)
                        key_list = line.split()
                        result_keys = self.validate(key_list)
                        cur_path = self.merge(result_keys, self.result)
                        if mandatory_options:
                            for option in mandatory_options.group(2).split():
                                mod_line = cur_path + ' ' + option
                                key_list = mod_line.split()
                                result_keys = self.validate(key_list)
                                self.merge(result_keys, self.result)
                    else:
                        if 'kiwi-ng --' in cur_path:
                            cur_path = ''
                        for mod_line in line.strip().split('|'):
                            mod_line = cur_path + ' ' + mod_line
                            mod_line = re.sub('[\[\]]', '', mod_line)
                            mod_line = re.sub('-h ', '', mod_line)
                            key_list = mod_line.split()
                            result_keys = self.validate(key_list)
                            self.merge(result_keys, self.result)

    def merge(self, key_list, result):
        raw_key_path = " ".join(key_list)
        key_path = ''
        for key in key_list:
            key_path += '[\'' + key + '\']'
        expression = 'self.result' + key_path
        exec(expression)
        return raw_key_path

    def validate(self, key_list):
        result_keys = []
        for key in key_list:
            option = re.search('^\[(--.*)=|^\[(.*)\]', key)
            mandatory = re.search('^(--.*)=', key)
            if option:
                if option.group(1):
                    result_keys.append(option.group(1))
            elif mandatory:
                if mandatory.group(1):
                    result_keys.append(mandatory.group(1))
            elif re.search('<servicename>|<command>|<args>...', key):
                pass
            else:
                key = key.replace('<', '__')
                key = key.replace('>', '__')
                result_keys.append(key)
        return result_keys


class AppTree:
    def __init__(self):
        self.completion = AppHash()
        self.level_dict = {}

    def traverse(self, tree=None, level=0, origin=None):
        if not tree:
            tree = self.completion.result['kiwi-ng']
        if not origin:
            origin = 'kiwi-ng'
        for key in tree:
            try:
                if self.level_dict[str(level)]:
                    pass
            except KeyError:
                self.level_dict[str(level)] = {}
            try:
                if self.level_dict[str(level)][origin]:
                    pass
            except KeyError:
                self.level_dict[str(level)][origin] = []

            if key not in self.level_dict[str(level)][origin]:
                self.level_dict[str(level)][origin].append(key)

            if tree[key]:
                self.traverse(tree[key], level + 1, key)


tree = AppTree()
tree.traverse()

# helpful for debugging
# pp = pprint.PrettyPrinter(indent=4)
# pp.pprint(tree.completion.result)

sorted_levels = collections.OrderedDict(
    sorted(tree.level_dict.items())
)

print(dedent('''
#========================================
# _kiwi
#----------------------------------------
function setupCompletionLine {
    local comp_line=$(echo $COMP_LINE | sed -e 's@kiwi-ng@kiwi@')
    local result_comp_line
    local prev_was_option=0
    for item in $comp_line; do
        if [ $prev_was_option = 1 ];then
            prev_was_option=0
            continue
        fi
        if [[ $item =~ -.* ]];then
            prev_was_option=1
            continue
        fi
        result_comp_line="$result_comp_line $item"
    done
    echo $result_comp_line
}

function _kiwi {
    local cur prev opts
    _get_comp_words_by_ref cur prev
    local cmd=$(setupCompletionLine | awk -F ' ' '{ print $NF }')
''').strip())

print('    for comp in $prev $cmd;do')
print('        case "$comp" in')
for level in sorted_levels:
    if level == '0':
        continue
    for sub in sorted(sorted_levels[level]):
        print('            "%s")' % (sub))
        print(
            '                __comp_reply "{0}"'.format(
                (" ".join(sorted(sorted_levels[level][sub])))
            )
        )
        print('                return 0')
        print('                ;;')
print('        esac')
print('    done')
print(
    '    __comp_reply "{0}"'.format(
        (" ".join(sorted(sorted_levels['0']['kiwi-ng'])))
    )
)
print('    return 0')
print('}')
print(dedent('''
#========================================
# comp_reply
#----------------------------------------
function __comp_reply {
    word_list=$@
    COMPREPLY=($(compgen -W "$word_list" -- ${cur}))
}

complete -F _kiwi -o default kiwi
complete -F _kiwi -o default kiwi-ng
''').strip())



import io
from unittest.mock import (
    patch, call, MagicMock, Mock
)
from pytest import raises
import unittest.mock as mock

from kiwi.bootloader.install.grub2 import BootLoaderInstallGrub2
from kiwi.defaults import Defaults

from kiwi.exceptions import (
    KiwiBootLoaderGrubInstallError,
    KiwiBootLoaderGrubDataError,
    KiwiBootLoaderGrubPlatformError,
    KiwiBootLoaderDiskPasswordError
)


class TestBootLoaderInstallGrub2:
    def setup(self):
        Defaults.set_platform_name('x86_64')

        self.firmware = mock.Mock()
        self.firmware.efi_mode = mock.Mock(
            return_value=None
        )

        self.custom_args = {
            'boot_device': '/dev/mapper/loop0p2',
            'root_device': '/dev/mapper/loop0p1',
            'efi_device': '/dev/mapper/loop0p3',
            'prep_device': '/dev/mapper/loop0p2',
            'system_root_volume': 'root',
            'system_volumes': {'boot/grub2': {
                'volume_options': 'subvol=@/boot/grub2',
                'volume_device': 'device'
            }},
            'firmware': self.firmware,
            'target_removable': None,
            'install_options': [],
            'shim_options': []
        }

        self.root_mount = mock.Mock()
        self.root_mount.device = self.custom_args['root_device']
        self.root_mount.mountpoint = 'tmp_root'

        self.volume_mount = mock.Mock()
        self.volume_mount.device = self.custom_args['root_device']
        self.volume_mount.mountpoint = 'tmp_volume'

        self.boot_mount = mock.Mock()
        self.boot_mount.device = self.custom_args['boot_device']
        self.boot_mount.mountpoint = 'tmp_boot'

        self.efi_mount = mock.Mock()
        self.efi_mount.device = self.custom_args['efi_device']
        self.efi_mount.mountpoint = 'tmp_efi'

        self.device_mount = mock.Mock()
        self.device_mount.device = 'devtmpfs'
        self.device_mount.mountpoint = 'dev'

        self.proc_mount = mock.Mock()
        self.proc_mount.device = 'proc'
        self.proc_mount.mountpoint = 'proc'

        self.sysfs_mount = mock.Mock()
        self.sysfs_mount.device = 'sysfs'
        self.sysfs_mount.mountpoint = 'sys'

        # the order of mount manager is reverse the order in the code
        self.mount_managers = [
            self.sysfs_mount,
            self.proc_mount,
            self.device_mount,
            self.volume_mount,
            self.efi_mount,
            self.boot_mount,
            self.root_mount
        ]

        device_provider = mock.Mock()
        device_provider.get_device = mock.Mock(
            return_value='/dev/some-device'
        )

        self.bootloader = BootLoaderInstallGrub2(
            Mock(), 'root_dir', device_provider, self.custom_args
        )

    def setup_method(self, cls):
        self.setup()

    @patch('kiwi.bootloader.install.grub2.Defaults.get_grub_path')
    def test_post_init_ppc_no_prep_device(self, mock_grub_path):
        self.bootloader.arch = 'ppc64'
        del self.custom_args['prep_device']
        with raises(KiwiBootLoaderGrubInstallError):
            self.bootloader.install()

    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    @patch('kiwi.bootloader.install.grub2.Defaults.get_grub_path')
    def test_grub2_bootloader_not_installed(
        self, mock_grub_path, mock_mount_manager, mock_command
    ):
        mock_grub_path.return_value = None
        self.bootloader.arch = 'x86_64'
        with raises(KiwiBootLoaderGrubDataError):
            self.bootloader.install()

    @patch('kiwi.bootloader.install.grub2.Defaults.get_grub_path')
    def test_unsupported_platform(self, mock_grub_path):
        self.bootloader.arch = 'unsupported'
        with raises(KiwiBootLoaderGrubPlatformError):
            self.bootloader.install()

    def test_post_init_no_boot_device(self):
        with raises(KiwiBootLoaderGrubInstallError):
            self.bootloader.post_init({})

    def test_post_init_no_root_device(self):
        with raises(KiwiBootLoaderGrubInstallError):
            self.bootloader.post_init({'boot_device': 'a'})

    def test_post_init_secure_boot_no_efi_device(self):
        self.firmware.efi_mode.return_value = 'uefi'
        del self.custom_args['efi_device']
        with raises(KiwiBootLoaderGrubInstallError):
            self.bootloader.post_init(self.custom_args)

    def test_install_required(self):
        assert self.bootloader.install_required() is True

    def test_install_required_ppc_opal(self):
        self.bootloader.arch = 'ppc64'
        self.bootloader.firmware = mock.Mock()
        self.bootloader.firmware.opal_mode = mock.Mock(
            return_value=True
        )
        assert self.bootloader.install_required() is False

    def test_install_required_efi_no_csm(self):
        self.bootloader.arch = 'x86_64'
        self.bootloader.firmware = mock.Mock()
        self.bootloader.firmware.efi_mode = mock.Mock(
            return_value='efi'
        )
        self.bootloader.firmware.legacy_bios_mode = mock.Mock(
            return_value=False
        )
        assert self.bootloader.install_required() is False

    def test_install_required_arm64(self):
        self.bootloader.arch = 'arm64'
        assert self.bootloader.install_required() is False

    def test_install_required_riscv64(self):
        self.bootloader.arch = 'riscv64'
        assert self.bootloader.install_required() is False

    @patch('kiwi.bootloader.install.grub2.Path.wipe')
    @patch('kiwi.bootloader.install.grub2.Path.which')
    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    @patch('kiwi.bootloader.install.grub2.Defaults.get_grub_path')
    @patch('kiwi.bootloader.install.grub2.glob.glob')
    @patch('kiwi.bootloader.install.grub2.SystemSetup')
    def test_install_with_extra_boot_partition(
        self, mock_SystemSetup, mock_glob, mock_grub_path,
        mock_mount_manager, mock_command, mock_which, mock_wipe
    ):
        setup = Mock()
        mock_SystemSetup.return_value = setup
        mock_glob.return_value = ['tmp_root/boot/grub2/grubenv']
        mock_grub_path.return_value = \
            self.root_mount.mountpoint + '/usr/lib/grub2/i386-pc'

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect

        self.bootloader.install()
        self.bootloader.root_mount.mount.assert_called_once_with(
            options=['subvol=root']
        )
        self.boot_mount.mount.assert_called_once_with()
        mock_glob.assert_called_once_with(
            'tmp_root/boot/*/grubenv'
        )
        mock_wipe.assert_called_once_with(
            'tmp_root/boot/grub2/grubenv'
        )
        mock_which.assert_called_once_with(
            root_dir='tmp_root', filename='grub2-install'
        )
        mock_command.assert_called_once_with(
            [
                'chroot', 'tmp_root', 'grub2-install', '--skip-fs-probe',
                '--directory', '/usr/lib/grub2/i386-pc',
                '--boot-directory', '/boot',
                '--target', 'i386-pc',
                '--modules', ' '.join(
                    Defaults.get_grub_bios_modules(multiboot=True)
                ),
                '/dev/some-device'
            ]
        )
        setup.setup_selinux_file_contexts.assert_called_once_with()

    @patch('kiwi.bootloader.install.grub2.Path.wipe')
    @patch('kiwi.bootloader.install.grub2.Path.which')
    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    @patch('kiwi.bootloader.install.grub2.Defaults.get_grub_path')
    @patch('kiwi.bootloader.install.grub2.glob.glob')
    @patch('os.path.exists')
    def test_install_on_dasd_disk(
        self, mock_os_path_exists, mock_glob, mock_grub_path,
        mock_mount_manager, mock_command, mock_which, mock_wipe
    ):
        mock_os_path_exists.return_value = True
        self.firmware.get_partition_table_type = mock.Mock(
            return_value='dasd'
        )

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect

        self.bootloader.install()
        self.bootloader.root_mount.mount.assert_called_once_with(
            options=['subvol=root']
        )
        self.boot_mount.mount.assert_called_once_with()
        assert mock_command.call_args_list == [
            call(
                [
                    'chroot', 'tmp_root', 'grub2-zipl-setup', '--keep'
                ]
            ),
            call(
                [
                    'mv',
                    'tmp_root/etc/default/zipl2grub.conf.in.orig',
                    'tmp_root/etc/default/zipl2grub.conf.in'
                ]
            ),
            call(
                [
                    'mv',
                    'tmp_root/boot/zipl/config',
                    'tmp_root/boot/zipl/config.kiwi'
                ]
            )
        ]

    @patch('kiwi.bootloader.install.grub2.Path.wipe')
    @patch('kiwi.bootloader.install.grub2.Path.which')
    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    @patch('kiwi.bootloader.install.grub2.Defaults.get_grub_path')
    @patch('kiwi.bootloader.install.grub2.glob.glob')
    @patch('kiwi.bootloader.install.grub2.SystemSetup')
    def test_install_ppc_ieee1275(
        self, mock_SystemSetup, mock_glob, mock_grub_path, mock_mount_manager,
        mock_command, mock_which, mock_wipe
    ):
        setup = Mock()
        mock_SystemSetup.return_value = setup
        mock_glob.return_value = ['tmp_root/boot/grub2/grubenv']
        mock_grub_path.return_value = \
            self.root_mount.mountpoint + '/usr/lib/grub2/powerpc-ieee1275'
        self.bootloader.arch = 'ppc64'

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect

        self.bootloader.install()
        self.root_mount.mount.assert_called_once_with(
            options=['subvol=root']
        )
        self.boot_mount.mount.assert_called_once_with()
        mock_wipe.assert_called_once_with(
            'tmp_root/boot/grub2/grubenv'
        )
        mock_command.assert_called_once_with(
            [
                'chroot', 'tmp_root', 'grub2-install', '--skip-fs-probe',
                '--no-nvram', '--directory', '/usr/lib/grub2/powerpc-ieee1275',
                '--boot-directory', '/boot',
                '--target', 'powerpc-ieee1275',
                '--modules', ' '.join(
                    Defaults.get_grub_ofw_modules()
                ),
                self.custom_args['prep_device']
            ]
        )
        setup.setup_selinux_file_contexts.assert_called_once_with()

    @patch('kiwi.bootloader.install.grub2.Path.wipe')
    @patch('kiwi.bootloader.install.grub2.Path.which')
    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    @patch('kiwi.bootloader.install.grub2.Defaults.get_grub_path')
    @patch('kiwi.bootloader.install.grub2.glob.glob')
    @patch('kiwi.bootloader.install.grub2.SystemSetup')
    def test_install_s390_emu(
        self, mock_SystemSetup, mock_glob, mock_grub_path, mock_mount_manager,
        mock_command, mock_which, mock_wipe
    ):
        setup = Mock()
        mock_SystemSetup.return_value = setup
        mock_glob.return_value = ['tmp_root/boot/grub2/grubenv']
        mock_grub_path.return_value = \
            self.root_mount.mountpoint + '/usr/lib/grub2/s390x-emu'
        self.bootloader.arch = 's390x'

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect

        self.bootloader.install()
        self.root_mount.mount.assert_called_once_with(
            options=['subvol=root']
        )
        self.boot_mount.mount.assert_called_once_with()
        mock_wipe.assert_called_once_with(
            'tmp_root/boot/grub2/grubenv'
        )
        mock_command.assert_called_once_with(
            [
                'chroot', 'tmp_root', 'grub2-install', '--skip-fs-probe',
                '--no-nvram', '--directory', '/usr/lib/grub2/s390x-emu',
                '--boot-directory', '/boot',
                '--target', 's390x-emu',
                '--modules', ' '.join(
                    Defaults.get_grub_s390_modules()
                ),
                '/dev/some-device'
            ]
        )
        setup.setup_selinux_file_contexts.assert_called_once_with()

    @patch('kiwi.bootloader.install.grub2.Path.wipe')
    @patch('kiwi.bootloader.install.grub2.Path.which')
    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    @patch('kiwi.bootloader.install.grub2.Defaults.get_grub_path')
    @patch('kiwi.bootloader.install.grub2.glob.glob')
    @patch('kiwi.bootloader.install.grub2.SystemSetup')
    def test_install(
        self, mock_SystemSetup, mock_glob, mock_grub_path, mock_mount_manager,
        mock_command, mock_which, mock_wipe
    ):
        setup = Mock()
        mock_SystemSetup.return_value = setup
        mock_which.return_value = None
        mock_glob.return_value = ['tmp_root/boot/grub2/grubenv']
        mock_grub_path.return_value = \
            self.root_mount.mountpoint + '/usr/lib/grub2/i386-pc'
        self.boot_mount.device = self.root_mount.device

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect
        self.bootloader.target_removable = True

        self.bootloader.install()
        self.root_mount.mount.assert_called_once_with(
            options=['subvol=root']
        )
        self.volume_mount.mount.assert_called_once_with(
            options=['subvol=@/boot/grub2']
        )
        mock_wipe.assert_called_once_with(
            'tmp_root/boot/grub2/grubenv'
        )
        mock_command.assert_called_once_with(
            [
                'chroot', 'tmp_root', 'grub2-install',
                '--removable', '--skip-fs-probe',
                '--directory', '/usr/lib/grub2/i386-pc',
                '--boot-directory', '/boot',
                '--target', 'i386-pc',
                '--modules', ' '.join(
                    Defaults.get_grub_bios_modules(multiboot=True)
                ),
                '/dev/some-device'
            ]
        )
        setup.setup_selinux_file_contexts.assert_called_once_with()

    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    @patch('os.path.exists')
    @patch('os.access')
    def test_secure_boot_install(
        self, mock_access, mock_exists,
        mock_mount_manager, mock_command
    ):
        mock_access.return_value = True
        mock_exists.return_value = True
        self.firmware.efi_mode.return_value = 'uefi'
        self.boot_mount.device = self.root_mount.device

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect

        self.bootloader.secure_boot_install()

        assert mock_command.call_args_list == [
            call([
                'chroot', 'tmp_root',
                'cp', '-p', '/usr/sbin/grub2-install',
                '/usr/sbin/grub2-install.orig'
            ]),
            call([
                'chroot', 'tmp_root',
                'cp', '/bin/true', '/usr/sbin/grub2-install'
            ]),
            call([
                'chroot', 'tmp_root', 'shim-install', '--removable',
                '/dev/some-device'
            ]),
            call([
                'chroot', 'tmp_root',
                'mv', '/usr/sbin/grub2-install.orig',
                '/usr/sbin/grub2-install'
            ])
        ]
        self.root_mount.mount.assert_called_once_with(
            options=['subvol=root']
        )
        self.volume_mount.mount.assert_called_once_with(
            options=['subvol=@/boot/grub2']
        )
        self.device_mount.bind_mount.assert_called_once_with()
        self.proc_mount.bind_mount.assert_called_once_with()
        self.sysfs_mount.bind_mount.assert_called_once_with()
        self.efi_mount.mount.assert_called_once_with()

    @patch('kiwi.bootloader.install.grub2.Path.which')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    def test_secure_boot_install_no_shim_install(
        self, mock_mount_manager, mock_which
    ):
        mock_which.return_value = None
        self.firmware.efi_mode.return_value = 'uefi'

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect

        self.bootloader.secure_boot_install()
        self.root_mount.mount.assert_called_once_with(
            options=['subvol=root']
        )
        self.volume_mount.mount.assert_called_once_with(
            options=['subvol=@/boot/grub2']
        )
        self.device_mount.bind_mount.assert_called_once_with()
        self.proc_mount.bind_mount.assert_called_once_with()
        self.sysfs_mount.bind_mount.assert_called_once_with()
        self.efi_mount.mount.assert_called_once_with()
        mock_which.assert_called_once_with(
            filename='shim-install', root_dir='tmp_root'
        )

    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    def test_set_disk_password(self, mock_mount_manager, mock_command):
        self.boot_mount.device = self.root_mount.device

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect

        self.bootloader.root_mount = MagicMock()
        self.bootloader.root_mount.mountpoint = 'root_mountpoint'
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            file_handle.read.return_value = 'data__cryptomount__data'
            self.bootloader.set_disk_password('credentials')
            file_handle.write.assert_called_once_with(
                'data__cryptomount -p "credentials"__data'
            )

    @patch('kiwi.bootloader.install.grub2.Command.run')
    @patch('kiwi.bootloader.install.grub2.MountManager')
    @patch('kiwi.bootloader.install.grub2.ExitStack')
    def test_set_disk_password_failed_to_open_grub_config(
        self, mock_ExitStack, mock_mount_manager, mock_command
    ):
        self.boot_mount.device = self.root_mount.device

        def side_effect(device, mountpoint=None):
            return self.mount_managers.pop()

        mock_mount_manager.side_effect = side_effect
        self.bootloader.root_mount = MagicMock()
        self.bootloader.root_mount.mountpoint = 'root_mountpoint'

        with patch('builtins.open', create=True) as mock_open:
            mock_open.side_effect = IOError()
            with raises(KiwiBootLoaderDiskPasswordError):
                self.bootloader.set_disk_password('credentials')



import io
from unittest.mock import (
    patch, call, MagicMock, Mock
)
from pytest import raises

from kiwi.storage.clone_device import CloneDevice

from kiwi.exceptions import KiwiRaidSetupError


class TestCloneDevice:
    def setup(self):
        self.storage_device = Mock()
        self.storage_device.get_device = Mock(
            return_value='/dev/source-device'
        )
        self.target_device = Mock()
        self.target_device.get_device = Mock(
            return_value='/dev/target-device'
        )
        self.clone_id = Mock()
        self.clone_device = CloneDevice(
            self.storage_device, 'root_dir'
        )

    def setup_method(self, cls):
        self.setup()

    @patch('kiwi.storage.clone_device.Command.run')
    @patch('kiwi.storage.clone_device.BlockID')
    @patch('kiwi.storage.clone_device.FileSystem.new')
    def test_clone_filesystem(
        self, mock_FileSystem_new, mock_BlockID, mock_Command_run
    ):
        self.clone_id.get_filesystem.return_value = 'ext3'
        mock_BlockID.return_value = self.clone_id

        self.clone_device.clone([self.target_device])

        mock_Command_run.assert_called_once_with(
            ['dd', 'if=/dev/source-device', 'of=/dev/target-device']
        )
        mock_FileSystem_new.assert_called_once_with(
            'ext3', self.target_device
        )
        mock_FileSystem_new.return_value.__enter__ \
            .return_value.set_uuid.assert_called_once_with()

    @patch('kiwi.storage.clone_device.Command.run')
    @patch('kiwi.storage.clone_device.BlockID')
    def test_clone_lvm(self, mock_BlockID, mock_Command_run):
        self.clone_id.get_filesystem.return_value = 'LVM2_member'
        mock_BlockID.return_value = self.clone_id

        self.clone_device.clone([self.target_device])

        assert mock_Command_run.call_args_list == [
            call(
                ['dd', 'if=/dev/source-device', 'of=/dev/target-device']
            ),
            call(
                ['vgimportclone', '/dev/target-device']
            )
        ]

    @patch('kiwi.storage.clone_device.Command.run')
    @patch('kiwi.storage.clone_device.BlockID')
    @patch('uuid.uuid4')
    def test_clone_luks(self, mock_uuid4, mock_BlockID, mock_Command_run):
        self.clone_id.get_filesystem.return_value = 'crypto_LUKS'
        mock_BlockID.return_value = self.clone_id
        mock_uuid4.return_value = 'some-UUID'

        self.clone_device.clone([self.target_device])

        assert mock_Command_run.call_args_list == [
            call(
                ['dd', 'if=/dev/source-device', 'of=/dev/target-device']
            ),
            call(
                [
                    'cryptsetup', '-q', 'luksUUID',
                    '/dev/target-device', '--uuid', 'some-UUID'
                ]
            )
        ]

    @patch('kiwi.storage.clone_device.Command.run')
    @patch('kiwi.storage.clone_device.BlockID')
    @patch('kiwi.storage.clone_device.FileSystem.new')
    @patch('kiwi.storage.clone_device.MappedDevice')
    def test_clone_raid(
        self, mock_MappedDevice, mock_FileSystem_new,
        mock_BlockID, mock_Command_run
    ):
        self.clone_id.get_filesystem.return_value = 'linux_raid_member'
        mock_BlockID.return_value = self.clone_id
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            file_handle.readline.return_value = \
                'ARRAY /dev/md0 metadata=1.2 name=asterix:0 UUID=...'

            self.clone_device.clone([self.target_device])

            mock_open.assert_called_once_with('root_dir/etc/mdadm.conf')

        mock_FileSystem_new.assert_called_once_with(
            mock_BlockID.return_value.get_filesystem.return_value,
            mock_MappedDevice.return_value
        )
        mock_FileSystem_new.return_value.__enter__ \
            .return_value.set_uuid.assert_called_once_with()
        assert mock_Command_run.call_args_list == [
            call(
                ['dd', 'if=/dev/source-device', 'of=/dev/target-device']
            ),
            call(
                ['mdadm', '--stop', '/dev/md0']
            ),
            call(
                [
                    'mdadm', '--assemble', '--update=uuid',
                    '--name', 'asterix:0', '/dev/md0', '/dev/target-device'
                ]
            ),
            call(
                ['mdadm', '--stop', '/dev/md0']
            ),
            call(
                ['mdadm', '--assemble', '/dev/md0', '/dev/source-device']
            )
        ]

    @patch('kiwi.storage.clone_device.Command.run')
    @patch('kiwi.storage.clone_device.BlockID')
    def test_clone_raid_raises(self, mock_BlockID, mock_Command_run):
        self.clone_id.get_filesystem.return_value = 'linux_raid_member'
        mock_BlockID.return_value = self.clone_id
        with patch('builtins.open', create=True) as mock_open:
            mock_open.side_effect = Exception
            with raises(KiwiRaidSetupError):
                self.clone_device.clone([self.target_device])
